module("L_HoneywellLib_webapi", package.seeall)

--------------------------------------------------------------------------------
--
-- Honeywell Total Connect Comfort backend for WebAPI API. This support is taken
-- from work in the Evohome Web Service.
--
-- Creator:  Mikee

-- Version 1.0: Dec 27, 2015 Mikee
--   Initial non working cut. Will need an account to further the development.

local main = { }

-- Upcalls
local debug = nil
local trace = nil
local log = nil
local err = nil
local log_http = nil
local updateCookies = nil
local createCookie = nil

local json = require("L_HoneywellTCC_json")
local ssl = require "ssl" -- 2020-02-14
local https = require("ssl.https")
local K = require("L_HoneywellTCC_const")

--------------------------------------------------------------------------------
-- Constants

-- Honeywell Web portal constants
local K_HOST = "rs.alarmnet.com"
local K_URL = "https://" .. K_HOST .. ":443/TotalConnectComfort/WebAPI/"
local K_USER_AGENT = "Apache-HttpClient/UNAVAILABLE (java 1.4)"

local LUASEC_PROTO = (ssl._VERSION or "0.5"):match("^0.5") and "tlsv1_2" or "any" -- rigpapa 2020-02-14

--------------------------------------------------------------------------------
-- Globals
local Session = 0
local TccDevice = 0
local UserId = 0
local RetryLimit = 1
local Cookies = { }

--------------------------------------------------------------------------------
-- Local functions

--++
-- Make an http request with retry
--
-- @param spec_options - table of additional or override options
-- @return res - 
--         code - https result code
--         header - response headers
--
function https_request(spec_options)
   trace("https_request", { url = spec_options.url })

   local res, code, header 
   local base_options = {
      ["headers"] = {
	 ["Accept"] = "*/*",
--	 ["Accept-Encoding"] = "gzip,deflate,sdch",
	 ["Accept-Encoding"] = "sdch",
	 ["Accept-Language"] = "en-US",
	 ["Cache-Control"] = "max-age=0",
	 ["Connection"] = "keep-alive",
	 ["Host"] = K_HOST,
      },
   }

   for i=1,RetryLimit do
      local options = { }
      options.headers = { }
      for name,value in pairs(base_options) do
	 if name == "headers" then
	    for hname, hvalue in pairs(value) do
	       options.headers[hname] = hvalue
	    end
	 else
	    options[name] = value
	 end
      end
      for name,value in pairs(spec_options) do
	 if name == "headers" then
	    for hname, hvalue in pairs(value) do
	       options.headers[hname] = hvalue
	    end
	 else
	    options[name] = value
	 end
      end

      local cookie = createCookie(Cookies)
      if cookie ~= "" then
	 options.headers["Cookie"] = cookie
      end

      -- Fix for LuaSec 0.8 and up, rigpapa 2020-02-14
      options.protocol = LUASEC_PROTO

      log_http("https req", { options=options })
      res, code, header = https.request(options)
      log_http("https resp", { res=res, code=code, header=header })

      if code ~= 400 and code ~= 504 then
	 break
      end
      trace("https_request retry", { res=res, code=code })
   end

   trace("https_request return", { res=res, code=code })
   return res, code, header
end

--++
-- Login to the Honeywell Total Connect Comfort website
--
-- @param user - username
-- @param pass - password
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--         token - session token
-- ------------------------------------------------------------------------------
local function login(user, pass)
   local time = os.time();

   local args = {
      ["ApplicationID"] = '91db1612-73fd-4500-91b2-e63b069b185c',
      ["Username"] = user,
      ["Password"] = pass,
   }

   local url = K_URL .. "api/Session"

   local body = json.encode(args)
   debug("body = " .. body)

   local result = { }
   local res, code, header = https_request {
      ["method"] = "POST",
      ["url"] = url,
      ["headers"] = {
	 ["content-type"] = "application/json",
	 ["content-length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   last_login_attempt = os.time();

   local status, reason
   if (code ~= 200) then
      reason = "Login failed, code = " .. code
      if code == 401 then
	 status = K.STATUS_AUTH_FAIL
      else
	 status = K.STATUS_FAIL
      end
   else
      debug("result = " .. table.concat(result))
      local res = json.decode(table.concat(result))
      status = K.STATUS_OK
      reason = "Success"

      Session = res.sessionId
      UserId = res.userInfo.userID
      log("login res", { sess=Session, user=UserId })
   end

   return status, reason, Session
end

--++
-- Get the list of devices per location
--
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--         locations - table of locations (See GetLocations for description)
--
local function get_locations()
   trace("get_locations")

   local status = K.STATUS_FAIL
   local reason = "Unknown failure"

   local url = K_URL .. "api/locations?userId=" .. UserId .. "&allData=True"
   local result = { }
   local res, code, header = https_request {
      ["method"] = "GET",
      ["url"] = url,
      ["headers"] = {
	 ["sessionId"] = Session,
      },
      ["sink"] = ltn12.sink.table(result),
   }

   if (code ~= 200) then
      reason = "GetLocations failed, code = " .. code
      status = K.STATUS_FAIL
      trace("get_locations return", { status=status, reason=reason })
      return status, reason, nil
   end

   debug("result = " .. table.concat(result))
   res = json.decode(table.concat(result))
   log("res", { res = res })

   local devices = { }
   for _,loc in ipairs(res) do

      local tbl = { }
      if loc.devices ~= nil then
	     for _,dev in ipairs(loc.devices) do
		tbl[loc.locationID .. ":" .. dev.deviceID] = { ["Name"] = dev.name }
	     end
      else
	 log("res.devices is nil")
      end
      devices[loc.locationID] = { }
      devices[loc.locationID].Devices = tbl
      devices[loc.locationID].Name = loc.name
   end

   status = K.STATUS_OK
   reason = "Success"

   trace("get_locations return", { status=status, reason=reason })
   return status, reason, devices
end

function string_to_hold_type(str)
      if str == "Scheduled" then
	 return K.STATUS_TYPE_SCHED_ID
      elseif str == "Temporary" then
	 return K.STATUS_TYPE_TEMP_ID
      elseif str == "Hold" then
	 return K.STATUS_TYPE_PERM_ID
      end
      return K.STATUS_TYPE_SCHED_ID
end

--++
-- Get the status of a thermostat
--
-- @param device_id = TCC device id
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--         stat - table of thermostat data (see GetStatus for description)
--
local function get_status(device_id)
   trace("get_status")

   local status = K.STATUS_FAIL
   local reason = "Unknown failure"

   local locationID, deviceID = string.match(device_id, "([^:]-):(.*)")

   local url = K_URL .. "api/locations?userId=" .. UserId .. "&allData=True"
   local result = { }
   local res, code, header = https_request {
      ["method"] = "GET",
      ["url"] = url,
      ["headers"] = {
	 ["sessionId"] = Session,
      },
      ["sink"] = ltn12.sink.table(result),
   }

   if (code ~= 200) then
      reason = "GetLocations failed, code = " .. code
      status = K.STATUS_FAIL
      trace("get_status return", { status=status, reason=reason })
      return status, reason, nil
   end

   debug("result = " .. table.concat(result))
   res = json.decode(table.concat(result))
   log_http("res", { res = res })

   local function find_location(locs, id)
      trace("find_location", { id=id })
      for _,loc in ipairs(locs) do
	 if (tostring(loc.locationID) == tostring(id)) then
	    trace("find_location return") --, { loc=loc })
	    return loc
	 end
      end
      trace("find_location nil")
      return nil
   end

   local function find_device(loc, id)
      trace("find_device", { id=id })
      for _,dev in ipairs(loc.devices) do
	 if (tostring(dev.deviceID) == tostring(id)) then
	    trace("find_device return") --, { dev=dev })
	    return dev
	 end
      end
      trace("find_device nil")
      return nil
   end

   local dev
   local loc = find_location(res, locationID)
   if loc ~= nil then
      dev = find_device(loc, deviceID)
   end

   if dev == nil then
      status = K.STATUS_FAIL
      reason = "Failed to find device " .. device_id
      trace("get_status return", { status=status, reason=reason })
      return status, reason, nil
   end

   local data = { }
   status = K.STATUS_OK
   reason = "Success"

   data.Success = true
   data.Load = 0

   data.CoolNextPeriod = 0
   data.CoolSetpoint = 0
   data.StatusCool = string_to_hold_type(dev.thermostat.changeableValues.status)
   if dev.thermostat.changeableValues.coolSetpoint ~= nil then
      data.CoolSetpoint = dev.thermostat.changeableValues.coolSetpoint.value
      if dev.thermostat.changeableValues.coolSetpoint.status ~= nil then
	 data.StatusCool = string_to_hold_type(dev.thermostat.changeableValues.coolSetpoint.status)
      end
   end

   data.DisplayUnits = (dev.thermostat.units == "Fahrenheit" and "F" or "C")
   data.DisplayTemp = dev.thermostat.indoorTemperature

   data.HeatNextPeriod = 0
   data.HeatSetpoint = 0
   data.StatusHeat = string_to_hold_type(dev.thermostat.changeableValues.status)
   if dev.thermostat.changeableValues.heatSetpoint ~= nil then
      data.HeatSetpoint = dev.thermostat.changeableValues.heatSetpoint.value
      if dev.thermostat.changeableValues.heatSetpoint.status ~= nil then
	 data.StatusHeat = string_to_hold_type(dev.thermostat.changeableValues.heatSetpoint.status)
      end
   end

   data.IndoorHumidity = dev.thermostat.indoorHumidity
   data.HasIndoorHumidity =
      (dev.thermostat.indoorHumidityStatus == "Measured" and true or false)

   data.OutdoorHumidity = dev.thermostat.outdoorHumidity
   data.HasOutdoorHumidity =
      (dev.thermostat.outdoorHumidityStatus == "Measured" and true or false)

   data.OutdoorTemp = dev.thermostat.outdoorTemperature
   data.HasOutdoorTemp =
      (dev.thermostat.outdoorTemperatureStatus == "Measured" and true or false)

   data.SchedCoolSetpoint = dev.thermostat.scheduleCoolSp
   data.SchedHeatSetpoint = dev.thermostat.scheduleCoolSp

   data.CanSwitchAuto = true
   local sysmode = dev.thermostat.changeableValues.mode
   if sysmode == "Off" then
      data.SystemSwitchPosition = K.SYSTEM_SWITCH_OFF_ID;
   else
      data.SystemSwitchPosition = K.SYSTEM_SWITCH_HEAT_ID;
   end

--   local fanmode = dev.settings.fan.changeableValues.mode
   data.FanSwitchPosition = 0
--      data.FanSwitchPosition = K.FAN_MODE_MAP1[fanmode]
   data.HasFan = 0

   trace("get_status return", { status=status, reason=reason })
   return status, reason, data
end

--++
-- Obtain schedule from the website for specified thermostat
--
-- @param device_id - TCC device id
--
local function get_schedule(device_id)
   return K.STATUS_FAIL, "Not implemented", nil
end

--++
-- Set thermostat values
--
-- @param device_id - TCC device ID
-- @param req - Requested state change (See SetStatus for description
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--
local function set_status(device_id, req)
   trace("set_status")

   local status = K.STATUS_FAIL
   local reason = "Unknown failure"

   local locationID, deviceID = string.match(device_id, "([^:]-):(.*)")

   function setit(url, body)
      trace("setit", { body=body })
      local body = json.encode(body)
      log("json body", { json=body })
      local result = { }
      local res, code, header = https_request {
	 ["method"] = "PUT",
	 ["url"] = url,
	 ["headers"] = {
	    ["sessionId"] = Session,
	    ["Content-Type"] = "application/json",
	    ["Content-Length"] = #body
	 },
	 ["source"] = ltn12.source.string(body),
	 ["sink"] = ltn12.sink.table(result),
      }
      trace("setit return", { res=res, code=code, header=header, result=result })
      return res, code, header, result
   end

   if req.HeatSetpoint ~= nil then
      local body = { }
      body["Value"] = req.HeatSetpoint
      if req.StatusHeat == K.STATUS_TYPE_PERM_ID then
	 body["Status"] = "Hold"
-- API will not accept this
	 body["NextTime"] = json.null
--	 body["NextTime"] = os.date("!%Y-%m-%dT%H:%M:%SZ", os.time() + 4*60*60) -- hold for 4 hours
      elseif req.StatusHeat == K.STATUS_TYPE_TEMP_ID then
	 body["Status"] = "Temporary"
	 body["NextTime"] = os.date("!%Y-%m-%dT%H:%M:%SZ", os.time() + 4*60*60) -- hold for 4 hours
      end

      local url = K_URL .. "api/devices/" .. deviceID ..
	 "/thermostat/changeableValues/heatSetpoint"

      local _,code = setit(url, body)
      if code ~= 200 and code ~= 201 then
	 return K.STATUS_HTTP_FAIL, "HTTP error " .. code
      end
   end

   if req.CoolSetpoint ~= nil then
      local body = { }
      if req.StatusCool == K.STATUS_TYPE_PERM_ID then
	 body["Value"] = req.CoolSetpoint
	 body["Status"] = "Hold"
      elseif req.StatusCool == K.STATUS_TYPE_TEMP_ID then
	 body["Value"] = req.CoolSetpoint
	 body["Status"] = "Temporary"
	 body["NextTime"] = "YYYY-mm-ddT23:00:00Z" -- XXX fix the date
      end

      local url = K_URL .. "api/devices/" .. deviceID ..
	 "/thermostat/changeableValues/coolSetpoint"

      local _,code = setit(url, body)
      if code ~= 200 and code ~= 201 then
	 return K.STATUS_HTTP_FAIL, "HTTP error " .. code
      end
   end

   if req.SystemSwitch ~= nil then
      local body = { }
--      body = K.SYSTEM_SWITCH_RMAP1[req.SystemSwitch]
      if req.SystemSwitch == K.SYSTEM_SWITCH_HEAT_ID then
	 body = "Heat"
      elseif req.SystemSwitch == K.SYSTEM_SWITCH_OFF_ID then
	 body = "Off"
      else
	 body = "None"
      end

      local url = K_URL .. "api/devices/" .. deviceID ..
	 "/thermostat/changeableValues/mode"

      local _,code = setit(url, body)
      if code ~= 200 and code ~= 201 then
	 return K.STATUS_HTTP_FAIL, "HTTP error " .. code
      end
   end

   status = K.STATUS_OK
   reason = "Success"

   return status, reason
end

--------------------------------------------------------------------------------
-- APIs

--++
-- Perform initialization
--
function Init(devId, imports)
   TccDevice = devId
   main = imports

   debug = main.log_http
   trace = main.trace
   log = main.log
   log_http = main.log_http
   err = main.err

   updateCookies = main.updateCookies
   createCookie = main.createCookie
end

--++
-- Login and return whatever token is reuired for subsequent calls. This token
-- will be passed on all library calls
--
-- @param user - string username
-- @param pass - string password
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         token - token to use on subsequent calls
function Login(user, pass)
   trace("WebAPI.Login", { user=user, pass=pass })

   local status, reason, session = login(user, pass)

   trace("WebAPI.Login return", { status=status, reason=reason, session=session })
   return status, reason, session
end

--++
-- Return a structure of location/device information
--
-- @param token - authentication token
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         devices {
--                  { locationID = {
--                    Name = "",  - string location name
--                    Devices = {
--                     { deviceId = { Name = name }},...
--                    }
--                   },...
--                  }
function GetDevices(token)
   trace("WebAPI.GetDevices", { token=token })

   Session = token
   if Session == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason, devices = get_locations()

   trace("WebAPI.GetDevices return", { status=status, reason=reason, devices=devices })
   return status, reason, devices
end

--++
-- Return the status of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         status - see L_HoneywellTCC_const for Get Status record
--
function GetStatus(token, deviceId)
   trace("WebAPI.GetStatus", { token=token, deviceId=deviceId })

   Session = token
   if Session == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason, info = get_status(deviceId)

   trace("WebAPI.GetStatus return", { status=status, reason=reason, info=info })
   return status, reason, info
end

--++
-- Set the operating parameters of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @param req - see L_HoneywellTCC_const for Set Status record
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--
function SetStatus(token, deviceId, req)
   trace("WebAPI.SetStatus", { token=token, deviceId=deviceId })

   Session = token
   if Session == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason = set_status(deviceId, req)

   trace("WebAPI.SetStatus return", { status=status, reason=reason })
   return status, reason
end
